#!/usr/bin/env python

from abc import ABC, abstractmethod
from typing import Any, Dict, Sequence, Type, Tuple, Union
from .candidates import Candidates
from .config import Config
from .info import InfoStyle
from .line_editor import LineEditor
from .screen import Screen
from .ansi import (  # noqa
    NL,
    SPACE,
    ERASE_LINE,
    BLACK,
    RED,
    GREEN,
    YELLOW,
    BLUE,
    PURPLE,
    CYAN,
    WHITE,
    BLACK_BG,
    RED_BG,
    GREEN_BG,
    YELLOW_BG,
    BLUE_BG,
    PURPLE_BG,
    CYAN_BG,
    WHITE_BG,
    RESET,
    BOLD,
    NEGATIVE,
)

__all__ = ["Layout", "DefaultLayout", "ReverseLayout", "ReverseListLayout", "get_layout", "list_layouts"]


layouts: Dict[str, Type["Layout"]] = {}


class Layout(ABC):
    def __init__(self, config: Config, candidates: Candidates):
        """
        Abstract layout

        Args:
            config: Configuration
            candidates: Candidates

        Attributes:
            config: Configuration
            candidates: Candidates
            offset: Screen offset
            line_editor: Line editor
            screen: Screen instance
        """
        self.config = config
        self.candidates = candidates
        self.offset: int = 0
        self.screen: Screen = None

    def __init_subclass__(cls, option: str, **kwargs: Dict[str, Any]) -> None:
        "Register a subclass"
        super().__init_subclass__(**kwargs)
        layouts[option] = cls

    @abstractmethod
    def update_screen(self, selected: int, erase: bool = True) -> None:
        "Update the screen - erase the old items, print the filtered items and the prompt"
        pass  # pragma: no cover

    @abstractmethod
    def move_selection(self, selected: int, lines: int = 0, pages: int = 0) -> int:
        "Move selection up/down n lines/pages"
        pass  # pragma: no cover

    @abstractmethod
    def enumerate_items(self) -> Sequence[Tuple[int, Any]]:
        "Return a sequence of enumerated item to by displayed"
        pass  # pragma: no cover

    @abstractmethod
    def clear_screen(self, erase: bool) -> None:
        "Clean the screen"
        pass  # pragma: no cover

    def cleanup(self) -> None:
        "Clean the screen when the finder is closed"
        if self.screen is not None:
            self.clear_screen(erase=True)
            self.screen.flush()

    @property
    def max_candidates_lines(self) -> int:
        "Maximun number of candidates printables on the screen"
        return self.screen.height - self.config.margin_lines

    @property
    def screen_items(self) -> Sequence[Any]:
        "Candidates to be displayed on the screen"
        return self.candidates.matching_candidates[self.offset : self.offset + self.max_candidates_lines]

    @property
    def screen_items_len(self) -> int:
        "Number of items on the screen"
        return len(self.screen_items)

    def screen_setup(self, line_editor: LineEditor) -> None:
        "Calculate the required height and setup the screen"
        self.line_editor = line_editor
        height: int = (
            self.config.height if self.config.height is not None else self.candidates.candidates_len + self.config.margin_lines
        )
        self.screen = Screen(stream=self.config.output_stream, fullscreen=self.config.fullscreen, height=height)
        self.update_screen(selected=0, erase=False)

    def calculate_offset(self, selected: int) -> None:
        "Calculate the screen offset"
        if selected >= self.offset + self.max_candidates_lines:
            self.offset = selected - self.max_candidates_lines + 1
        elif selected < self.offset:
            self.offset = selected
        if self.offset < 0:
            self.offset = 0

    def print_header(self) -> None:
        "Print header"
        if self.config.header_str:
            self.screen.erase_line().write(self.config.header_str).nl()

    def print_items(self, selected: int) -> None:
        "Print candidates"
        for i, item in self.enumerate_items():
            is_selected = i + self.offset == selected
            self.screen.erase_line()
            if is_selected:
                self.screen.write(f"{RED}{BOLD}{BLACK_BG}{self.config.pointer_str}").reset().bold()
            else:
                self.screen.write(f"{BLACK_BG}{self.config.no_pointer_str}").reset()
            self.screen.space(1).write(self.config.format_fn(item)).reset().nl()

    def print_empty_lines(self, delta: int = 0) -> None:
        "Print empty lines"
        lines = self.max_candidates_lines - self.screen_items_len + delta
        self.screen.nl(lines)

    def print_info(self) -> None:
        "Print info"
        if self.config.info_style == InfoStyle.DEFAULT:
            matching_candidates_len = self.candidates.matching_candidates_len
            candidates_len = self.candidates.candidates_len
            self.screen.erase_line().space(2).write(f"{YELLOW}{matching_candidates_len}/{candidates_len}").reset().nl()

    def print_prompt(self) -> None:
        "Print prompt"
        self.screen.erase_line().write(f"{CYAN}{self.config.prompt_str}").reset()
        self.line_editor.print(self.screen)


class DefaultLayout(Layout, option="default"):
    "Display from the bottom of the screen"

    def update_screen(self, selected: int, erase: bool = True) -> None:
        "Update the screen - erase the old items, print the filtered items and the prompt"
        self.calculate_offset(selected)
        self.clear_screen(erase)
        self.print_header()
        self.print_empty_lines()
        self.print_items(selected)
        self.print_info()
        self.print_prompt()
        self.screen.flush()

    def enumerate_items(self) -> Sequence[Tuple[int, Any]]:
        "Return a sequence of enumerated item to by displayed"
        return reversed(list(enumerate(self.screen_items)))  # type: ignore

    def move_selection(self, selected: int, lines: int = 0, pages: int = 0) -> int:
        "Move selection up/down n lines"
        return selected + lines + pages * self.max_candidates_lines

    def clear_screen(self, erase: bool) -> None:
        "Clean the screen"
        if erase:
            lines = self.screen.height - 1
            self.screen.erase_line().move_up(lines).erase_lines(lines)


class ReverseLayout(Layout, option="reverse"):
    "Display from the top of the screen"

    def update_screen(self, selected: int, erase: bool = True) -> None:
        "Update the screen - erase the old items, print the filtered items and the prompt"
        self.calculate_offset(selected)
        self.clear_screen(erase)
        self.print_header()
        self.screen.nl()  # skip this line for the prompt
        self.print_info()
        self.print_items(selected)
        self.print_empty_lines(delta=-1 if self.config.fullscreen else 0)
        self.screen.move_up(self.max_candidates_lines + 1 + self.config.info_lines)  # jump back to the prompt line
        self.print_prompt()
        self.screen.flush()

    def enumerate_items(self) -> Sequence[Tuple[int, Any]]:
        "Return a sequence of enumerated item to by displayed"
        return list(enumerate(self.screen_items))

    def move_selection(self, selected: int, lines: int = 0, pages: int = 0) -> int:
        "Move selection up/down n lines"
        return selected - lines - pages * self.max_candidates_lines

    def clear_screen(self, erase: bool) -> None:
        "Clean the screen"
        if erase:
            lines: int = self.config.header_lines + self.config.prompt_lines - 2
            self.screen.move_up(lines)
            self.screen.erase_lines(self.screen.height)


class ReverseListLayout(ReverseLayout, option="reverse-list"):
    "Display from the top of the screen, prompt at the bottom"

    def update_screen(self, selected: int, erase: bool = True) -> None:
        "Update the screen - erase the old items, print the filtered items and the prompt"
        self.calculate_offset(selected)
        self.clear_screen(erase)
        self.print_header()
        self.print_items(selected)
        self.print_empty_lines()
        self.print_info()
        self.print_prompt()
        self.screen.flush()

    def clear_screen(self, erase: bool) -> None:
        "Clean the screen"
        if erase:
            lines = self.screen.height - 1
            self.screen.erase_line().move_up(lines).erase_lines(lines)


def get_layout(layout: Union[str, Type[Layout]], config: Config, candidates: Candidates) -> Layout:
    "Get a layout instance by name or by class"
    if isinstance(layout, str):
        return layouts[layout](config, candidates)
    else:
        return layout(config, candidates)


def list_layouts() -> Sequence[str]:
    "List layouts"
    return list(layouts.keys())



#!/usr/bin/env python

from .finder import (
    Finder,
    DEFAULT_POINTER,
    DEFAULT_PROMPT,
    DEFAULT_HEADER,
    DEFAULT_LAYOUT,
    DEFAULT_MATCHER,
)
from .keys import KeysBinding
from .matcher import Matcher
from .layout import Layout
from .info import InfoStyle
from .exceptions import AcceptAction, AbortAction, CustomAction, GenericAction
from .prompt import Prompt
from typing import Any, Callable, Iterator, Optional, Sequence, Set, Type, Union

__version__ = "0.0.22"
"PZP Version"

__all__ = [
    "pzp",
    "CustomAction",
    "GenericAction",
    "Finder",
]


def pzp(
    candidates: Union[Callable[[], Sequence[Any]], Iterator[Any], Sequence[Any]],
    height: Optional[int] = None,
    fullscreen: bool = True,
    format_fn: Callable[[Any], str] = lambda x: str(x),
    layout: Union[Type[Layout], str] = DEFAULT_LAYOUT,
    info_style: InfoStyle = InfoStyle.DEFAULT,
    pointer_str: str = DEFAULT_POINTER,
    prompt_str: str = DEFAULT_PROMPT,
    header_str: str = DEFAULT_HEADER,
    keys_binding: Optional[KeysBinding] = None,
    matcher: Union[Matcher, Type[Matcher], str] = DEFAULT_MATCHER,
    lazy: bool = False,
    handle_actions: Set[Type[GenericAction]] = {AcceptAction, AbortAction},
    input: Optional[str] = None,
) -> Any:
    """
    Open pzp and return the selected element

    If the Lazy mode is enabled, starts the finder only if the candidates are more than one.
    If there is only one match returns the only match, if there is no match returns None.

    Examples:
        >>> pzp(candidates=list(Path('.').iterdir()))
        PosixPath('README.md')

    Args:
        candidates: Candidates
        fullscreen: Full screen mode
        height: Finder window height
        format_fn: Items format function
        layout: Finder layout
        info_style: Determines the display style of finder info
        pointer_str: Pointer to the current line
        prompt_str: Input prompt
        header_str: Header
        keys_binding: Custom keys binding
        matcher: Matcher
        lazy: Lazy mode, starts the finder only if the candidates are more than one
        handle_actions: Actions to be handled

    Returns:
        item: the selected item
    """
    finder = Finder(
        candidates=candidates,
        fullscreen=fullscreen,
        height=height,
        format_fn=format_fn,
        layout=layout,
        info_style=info_style,
        pointer_str=pointer_str,
        prompt_str=prompt_str,
        header_str=header_str,
        keys_binding=keys_binding,
        matcher=matcher,
        lazy=lazy,
    )
    try:
        finder.show(input=input)
    except GenericAction as ex:
        if type(ex) in (handle_actions or set()):
            return ex.selected_item if isinstance(ex, AcceptAction) else None
        else:
            raise


def prompt(
    text: str = "",
    default: Optional[Any] = None,
    prompt_suffix: str = ": ",
    show_default: bool = True,
    type: Optional[Any] = None,
    value_proc: Optional[Callable[[str], Any]] = None,
    input: Optional[str] = None,
) -> Any:
    """
    Ask for user input.

    Examples:
        >>> prompt("Name")
        your input

    Args:
        text: Prompt text
        default: Default value to use if no input happens.
                 If this is not given it will prompt until it's aborted.
        prompt_suffix: Suffix that should be added to the prompt
        show_default: Show default value in the prompt
        type: The type to use to check the value against
        value_proc: Type conversion function

    Returns:
        item: the selected item
    """
    if default is not None and show_default:
        prompt_str = f"{text} [{default}]{prompt_suffix}".lstrip()
    else:
        prompt_str = f"{text}{prompt_suffix}"
    prompt = Prompt(prompt_str=prompt_str, default=default, type=type, value_proc=value_proc)
    while True:
        try:
            prompt.show(input=input)
        except AcceptAction as ex:
            if ex.selected_item:
                return ex.selected_item
            if default is not None:
                return default
        finally:
            print()



#!/usr/bin/env python

from typing import Any, Callable, Iterator, Sequence, Type, Union
from .matcher import Matcher, get_matcher

__all__ = [
    "Candidates",
]


class Candidates:
    def __init__(
        self,
        candidates: Union[Callable[[], Sequence[Any]], Iterator[Any], Sequence[Any]],
        format_fn: Callable[[Any], str],
        matcher: Union[Matcher, Type[Matcher], str],
    ):
        """
        Candidates

        Args:
            candidates: Candidates
            format_fn: Items format function
            matcher: Matcher

        Attributes:
            format_fn: Items format function
            matcher: Matcher
            candidates: candidates
            matching_candidates_len: filtered items
        """
        self.format_fn = format_fn
        self.matcher = get_matcher(matcher)
        # Get the candidates
        if isinstance(candidates, Iterator) or callable(candidates):
            self.get_items_fn: Union[None, Callable[[], Sequence[Any]], Iterator[Any]] = candidates
            self.candidates: Sequence[Any] = []
        else:
            self.get_items_fn = None
            self.candidates = candidates

    @property
    def candidates_len(self) -> int:
        "Number of candidates"
        return len(self.candidates)

    @property
    def matching_candidates_len(self) -> int:
        "Number of matching candidates"
        return len(self.matching_candidates)

    def refresh_candidates(self) -> None:
        "Load/reload the candidate list"
        # Get items
        if isinstance(self.get_items_fn, Iterator):
            self.candidates = list(self.get_items_fn)
        elif callable(self.get_items_fn):
            self.candidates = list(self.get_items_fn())

    def apply_filter(self, pattern: str) -> None:
        "Filter the items"
        self.matching_candidates: Sequence[Any] = self.matcher.filter(
            pattern=pattern, candidates=self.candidates, format_fn=self.format_fn
        )



#!/usr/bin/env python

import sys
from typing import Any, Callable, Optional, TextIO
from .actions import Action, ActionsHandler
from .exceptions import AcceptAction, AbortAction, CustomAction, MissingHander
from .keys import KeyEvent, KeysBinding
from .line_editor import LineEditor
from .screen import Screen

__all__ = [
    "Prompt",
]


class Prompt(ActionsHandler):
    def __init__(
        self,
        default: Optional[Any] = None,
        prompt_str: str = "",
        type: Optional[Any] = None,
        value_proc: Optional[Callable[[str], Any]] = None,
        keys_binding: Optional[KeysBinding] = None,
        output_stream: TextIO = sys.stderr,
    ):
        """
        Initializate Prompt object

        Args:
            default: The default value to use if no input happens
            prompt_str: The input prompt
            type: The type to use to check the value against
            value_proc: Type conversion function
            keys_binding: Custom keys binding
            output_stream: Output stream
        """
        super().__init__(keys_binding=keys_binding)
        self.default = default
        self.prompt_str = prompt_str
        self.output_stream = output_stream
        if value_proc is not None:
            self.value_proc: Optional[Callable[[str], Any]] = value_proc
        elif (type is not None) or (default is not None and not isinstance(default, str)):
            from click.types import convert_type

            self.value_proc = convert_type(type, default)
        else:
            self.value_proc = None

    def setup(self, input: Optional[str] = None) -> None:
        """
        Setup Prompt execution

        Args:
            input: initial string
        """
        self.line_editor = LineEditor(line=input or "", keys_handler=self.keys_handler)
        self.screen = Screen(stream=self.output_stream, fullscreen=False)
        self.update_screen()

    def show(self, input: Optional[str] = None) -> Any:
        """
        Open pzp and return the selected element

        Args:
            input: initial string

        Raises:
            AcceptAction: Raises when the user presses a key that is mapped to the "accept" action.
            AbortAction: Raises when the user presses a key that is mapped to the "abort" action.
            CustomAction: Raises when the user presses a key that is mapped to the "custom" action.
        """
        self.setup(input=input)
        while True:
            self.process_key()
            self.update_screen()

    def process_key(self, ch: Optional[str] = None) -> None:
        "Process the pressed key"
        key_event = self.keys_handler.get_key_event(ch)
        try:
            self.line_editor.process_key_event(key_event)
        except MissingHander:
            try:
                self.process_key_event(key_event)
            except MissingHander:
                raise CustomAction(action=key_event.action, ch=key_event.ch, selected_item=self.prepare_result(), line=self.line_editor.line)  # type: ignore

    def update_screen(self) -> None:
        "Update the screen"
        self.screen.write("\r").erase_line()
        self.screen.erase_line().write(self.prompt_str).reset()
        self.line_editor.print(self.screen)
        self.screen.flush()

    def prepare_result(self) -> Any:
        "Return the line converted to the correct type or the default value if the line is empty"
        if not self.line_editor.line:
            return self.default
        elif self.value_proc is not None:
            # value_proc raises an exception if the value is invalid
            return self.value_proc(self.line_editor.line)
        else:
            return self.line_editor.line

    @Action("accept", keys=["enter"])
    def accept(self, key_event: KeyEvent) -> None:
        "Confirm"
        try:
            selected_item = self.prepare_result()
        except Exception as ex:
            self.screen.write(f"\nError: The value you entered was invalid: {ex}\n")
            self.screen.flush()
            return
        raise AcceptAction(action="accept", ch=key_event.ch, selected_item=selected_item, line=self.line_editor.line)

    @Action("abort", keys=["ctrl-c", "ctrl-g", "ctrl-q", "esc"])
    def abort(self, key_event: KeyEvent) -> None:
        "Cancel"
        raise AbortAction(action="abort", ch=key_event.ch, line=self.line_editor.line)

    @Action("ignore", keys=["null", "insert"])
    def ignore(self) -> None:
        "Do nothing"
        pass



from .commons import prepare


def test_term_reverse_list_fullscreen():
    terminal, finder, k = prepare(fullscreen=True, layout="reverse-list")

    k("down")
    assert finder.selected == 1
    assert terminal.cursor.y == 23

    k("9")
    assert finder.selected == 1
    assert terminal.cursor.y == 23

    k("9")
    assert finder.selected == 0
    assert finder.prepare_result() == "0x99"
    assert terminal.cursor.y == 23

    k("x")
    assert finder.selected == 0
    assert finder.prepare_result() is None
    assert terminal.cursor.y == 23

    finder.layout.screen.cleanup()
    assert terminal.cursor.y == 0


def test_term_reverse_list_hidden():
    terminal, finder, k = prepare(fullscreen=False, layout="reverse-list", candidates=["Yes", "No"], info_style="hidden")
    output_stream = terminal.output_stream
    lines = 10
    for i in range(0, lines):
        output_stream.write(f"line{i}>\n")
    assert terminal.cursor.y == lines + 9

    k("down")
    assert finder.selected == 1
    assert terminal.cursor.y == lines + 9

    k("Y")
    assert finder.selected == 0
    assert terminal.cursor.y == lines + 9

    k("e")
    assert finder.selected == 0
    assert finder.prepare_result() == "Yes"
    assert terminal.cursor.y == lines + 9

    k("X")
    assert finder.selected == 0
    assert finder.prepare_result() is None
    assert terminal.cursor.y == lines + 9


def test_term_reverse_fullscreen():
    terminal, finder, k = prepare(fullscreen=True, layout="reverse")

    k("down")
    assert finder.selected == 1
    assert terminal.cursor.y == 0

    k("9")
    assert finder.selected == 1
    assert terminal.cursor.y == 0

    k("9")
    assert finder.selected == 0
    assert finder.prepare_result() == "0x99"
    assert terminal.cursor.y == 0

    k("x")
    assert finder.selected == 0
    assert finder.prepare_result() is None
    assert terminal.cursor.y == 0

    finder.layout.screen.cleanup()
    assert terminal.cursor.y == 0


def test_term_reverse():
    terminal, finder, k = prepare(fullscreen=False, layout="reverse", candidates=["Yes", "No"])
    output_stream = terminal.output_stream
    lines = 10
    for i in range(0, lines):
        output_stream.write(f"line{i}>\n")
    assert terminal.cursor.y == lines

    k("down")
    assert finder.selected == 1
    assert terminal.cursor.y == lines

    k("Y")
    assert finder.selected == 0
    assert terminal.cursor.y == lines

    k("e")
    assert finder.selected == 0
    assert finder.prepare_result() == "Yes"
    assert terminal.cursor.y == lines

    k("X")
    assert finder.selected == 0
    assert finder.prepare_result() is None
    assert terminal.cursor.y == lines


def test_term_reverse_hidden():
    terminal, finder, k = prepare(fullscreen=False, layout="reverse", candidates=["Yes", "No"], info_style="hidden")
    output_stream = terminal.output_stream
    lines = 10
    for i in range(0, lines):
        output_stream.write(f"line{i}>\n")
    assert terminal.cursor.y == lines

    k("down")
    assert finder.selected == 1
    assert terminal.cursor.y == lines

    k("Y")
    assert finder.selected == 0
    assert terminal.cursor.y == lines

    k("e")
    assert finder.selected == 0
    assert finder.prepare_result() == "Yes"
    assert terminal.cursor.y == lines

    k("X")
    assert finder.selected == 0
    assert finder.prepare_result() is None
    assert terminal.cursor.y == lines



from pzp.keys import key_to_str, KeysHandler


def test_key_to_str():
    t = key_to_str("a")
    assert t == "0x61"
    t = key_to_str("ctrl-o")
    assert t == "ctrl-o"


def test_keys_handler():
    keys_binding = {"custom": ["ctrl-o"], "qu-qu": ["ctrl-q"], "oh-oh": ["!"]}
    h = KeysHandler(keys_binding)
    ev = h.get_key_event("\x0f")
    assert ev.ch == "\x0f"
    assert ev.action == "custom"
    ev = h.get_key_event("a")
    assert ev.ch == "a"
    assert ev.action is None


def test_default_keys_handler():
    h = KeysHandler()
    ev = h.get_key_event("!")
    assert ev.ch == "!"
    assert ev.action is None
    h.set_keys_binding(["!", "$"], "custom")
    ev = h.get_key_event("!")
    assert ev.ch == "!"
    assert ev.action == "custom"
    ev = h.get_key_event("$")
    assert ev.ch == "$"
    assert ev.action == "custom"



#!/usr/bin/env python

from typing import Any, Callable, Optional, TextIO, Union
from .info import InfoStyle, get_style
from .ansi import (  # noqa
    NL,
    SPACE,
    ERASE_LINE,
    BLACK,
    RED,
    GREEN,
    YELLOW,
    BLUE,
    PURPLE,
    CYAN,
    WHITE,
    BLACK_BG,
    RED_BG,
    GREEN_BG,
    YELLOW_BG,
    BLUE_BG,
    PURPLE_BG,
    CYAN_BG,
    WHITE_BG,
    RESET,
    BOLD,
    NEGATIVE,
)

__all__ = ["Config"]


class Config:
    def __init__(
        self,
        fullscreen: bool,
        height: Optional[int],
        format_fn: Callable[[Any], str],
        info_style: Union[InfoStyle, str],
        pointer_str: str,
        prompt_str: str,
        header_str: str,
        lazy: bool,
        output_stream: TextIO,
    ):
        """
        Finder config

        Args:
            fullscreen: Full screen mode
            height: Finder window height
            format_fn: Items format function
            info_style: Determines the display style of finder info
            pointer_str: Pointer to the current line
            prompt_str: Input prompt
            header_str: Header
            lazy: Lazy mode, starts the finder only if the candidates are more than one
            output_stream: Output stream
        """
        self.fullscreen = fullscreen
        self.height = height
        self.format_fn = format_fn
        self.info_style: InfoStyle = get_style(info_style)
        self.pointer_str = pointer_str
        self.no_pointer_str = " " * len(pointer_str)
        self.prompt_str = prompt_str
        self.header_str = header_str
        self.lazy = lazy
        self.output_stream = output_stream

    @property
    def info_lines(self) -> int:
        "Number of info lines"
        return 1 if self.info_style == InfoStyle.DEFAULT else 0

    @property
    def prompt_lines(self) -> int:
        "Number of prompt lines"
        return len(self.prompt_str.split(f"{NL}"))

    @property
    def header_lines(self) -> int:
        "Number of header lines"
        return len(self.header_str.split(f"{NL}")) if self.header_str else 0

    @property
    def margin_lines(self) -> int:
        "Screen margin"
        return self.info_lines + self.prompt_lines + self.header_lines



#!/usr/bin/env python

from typing import Dict, Optional, Sequence
from .input import get_char, KEYS_MAPPING

__all__ = [
    "KEYS",
    "KeyEvent",
    "KeysHandler",
    "KeysBinding",
    "key_to_str",
]

KeysBinding = Dict[str, Sequence[str]]

KEYS = {k: k for k in KEYS_MAPPING.values()}
KEYS.update(
    {
        "ctrl-a": "\x01",
        "ctrl-b": "\x02",
        "ctrl-c": "\x03",
        "ctrl-d": "\x04",
        "ctrl-e": "\x05",
        "ctrl-f": "\x06",
        "ctrl-g": "\x07",
        "ctrl-h": "\x08",
        "ctrl-i": "\x09",
        "ctrl-j": "\x0a",
        "ctrl-k": "\x0b",
        "ctrl-l": "\x0c",
        "ctrl-m": "\x0d",
        "ctrl-n": "\x0e",
        "ctrl-o": "\x0f",
        "ctrl-p": "\x10",
        "ctrl-q": "\x11",
        "ctrl-r": "\x12",
        "ctrl-s": "\x13",
        "ctrl-t": "\x14",
        "ctrl-u": "\x15",
        "ctrl-v": "\x16",
        "ctrl-w": "\x17",
        "ctrl-x": "\x18",
        "ctrl-y": "\x19",
        "ctrl-z": "\x1a",
        "ctrl-\\": "\x1c",
        "ctrl-]": "\x1d",
        "ctrl-^": "\x1e",
        "ctrl-/": "\x1f",
        "space": " ",
        "null": "\0",
        "nl": "\n",
        "enter": "\r",
        "tab": "\t",
        "bspace": "\x7f",
        "bs": "\x7f",
        "esc": "\x1b",
        "shift-tab": "btab",
        "page-up": "pgup",
        "page-down": "pgdn",
    }
)


def key_to_str(ch: str) -> str:
    "Return the textual representation of a char"
    return f"0x{ord(ch):x}" if len(ch) == 1 else f"{ch}"


class KeyEvent:
    def __init__(self, ch: str, action: Optional[str]) -> None:
        """
        Key Event represents a key action on the keyboard.

        Args:
            ch: Pressed key
            action: Action

        Attributes:
            ch: Pressed key
            action: Action
        """
        self.ch = ch
        self.action = action

    def __str__(self) -> str:
        return f"<{key_to_str(self.ch)}, {self.action or '-'}>"  # pragma: no cover


class KeysHandler:
    def __init__(self, keys_binding: Optional[KeysBinding] = None) -> None:
        """
        Keys handler is a collection of bindings of keys to actions.

        Args:
            keys_binding: Custom key binding

        Attributes:
            keycodes_actions: key => action mapping
        """
        self.keycodes_actions: Dict[str, str] = {}
        if keys_binding:
            self.update(keys_binding)

    def update(self, keys_binding: KeysBinding) -> None:
        for action, keys in keys_binding.items():
            self.set_keys_binding(keys, action)

    def set_keys_binding(self, keys: Sequence[str], action: str) -> None:
        "Add a binding for one or more keys to an action"
        for key in keys:
            self.set_key_binding(key, action)

    def set_key_binding(self, key: str, action: str) -> None:
        "Add a binding for one key to an action"
        self.keycodes_actions[KEYS[key] if len(key) > 1 else key] = action

    def get_key_event(self, ch: Optional[str] = None) -> KeyEvent:
        if ch is None:
            ch = get_char()
        action = self.keycodes_actions.get(ch)
        return KeyEvent(ch=ch, action=action)

    def __str__(self) -> str:
        return "\n".join(f"<{key_to_str(ch)}, {action or '-'}>" for ch, action in self.keycodes_actions.items())  # pragma: no cover



from pzp.line_editor import LineEditor


def test_line_editor():
    le = LineEditor()
    assert len(le) == 0
    assert str(le) == ""
    t = "abcdefgh"
    for c in t:
        le.insert(c)
    assert str(le) == t
    assert len(le) == len(t)
    assert le.cursor_pos == len(t)
    le.delete_char()
    assert str(le) == t
    assert len(le) == len(t)
    le.delete_backward_char()
    assert str(le) == t[:-1]
    assert len(le) == len(t) - 1
    le.beginning_of_line()
    assert le.cursor_pos == 0
    le.delete_backward_char()
    assert str(le) == t[:-1]
    assert len(le) == len(t) - 1


def test_line_editor_move():
    t = "abcdefgh"
    le = LineEditor(t)
    le.set_cursor_pos(0)
    le.forward_char()
    le.forward_char()
    assert le.cursor_pos == 2
    for _ in range(0, 10):
        le.forward_char()
    assert le.cursor_pos == len(le)
    le.backward_char()
    le.backward_char()
    assert le.cursor_pos == len(le) - 2
    for _ in range(0, 20):
        le.backward_char()
    assert le.cursor_pos == 0
    le.end_of_line()
    assert le.cursor_pos == len(le)
    le.beginning_of_line()
    assert le.cursor_pos == 0



from enum import Enum
from typing import Sequence, Union

__all__ = [
    "InfoStyle",
    "list_styles",
    "get_style",
]


class InfoStyle(Enum):
    "Display style of finder info"

    DEFAULT = "default"
    " Display on the next line to the prompt "
    HIDDEN = "hidden"
    " Do not display finder info"


def list_styles() -> Sequence[str]:
    "List styles"
    return [InfoStyle.DEFAULT.value, InfoStyle.HIDDEN.value]


def get_style(style: Union[str, InfoStyle]) -> InfoStyle:
    "Get a style by name"
    if style == InfoStyle.DEFAULT:
        return InfoStyle.DEFAULT
    elif style == InfoStyle.HIDDEN:
        return InfoStyle.HIDDEN
    else:
        return InfoStyle.DEFAULT



import pytest
from pzp.finder import Finder
from pzp.keys import KEYS
from pzp.exceptions import AcceptAction, AbortAction, CustomAction


def test_actions():
    candidates = [f"0x{x}" for x in range(0, 100)]
    finder = Finder(
        candidates=candidates,
        height=10,
        fullscreen=True,
        layout="reverse-list",
        keys_binding={"custom": ["ctrl-o"], "ignore": ["ctrl-i"]},
    )
    finder.setup()
    assert finder.selected == 0

    finder.process_key("down")
    finder.apply_filter()
    finder.update_screen()
    assert finder.selected == 1

    finder.process_key(KEYS["pgup"])
    finder.apply_filter()
    assert finder.selected == 0

    finder.process_key(KEYS["pgdn"])
    finder.apply_filter()
    assert finder.selected == 22

    finder.process_key("up")
    finder.apply_filter()
    assert finder.selected == 21

    finder.process_key(KEYS["pgup"])
    finder.apply_filter()
    assert finder.selected == 0

    finder.process_key(KEYS["ctrl-a"])
    finder.process_key(KEYS["ctrl-b"])
    finder.process_key(KEYS["ctrl-i"])
    with pytest.raises(CustomAction):
        finder.process_key(KEYS["ctrl-o"])
    with pytest.raises(AbortAction):
        finder.process_key(KEYS["ctrl-c"])
    with pytest.raises(AcceptAction) as except_info:
        finder.process_key(KEYS["enter"])
    assert except_info.value.action == "accept"
    assert except_info.value.ch == KEYS["enter"]
    assert except_info.value.selected_item == "0x0"



#!/usr/bin/env python

import inspect
from typing import Any, Callable, Dict, Optional, Sequence, Union
from .keys import KeyEvent, KeysHandler, KeysBinding
from .exceptions import MissingHander

__all__ = [
    "Action",
    "ActionsHandler",
]

ActionHandler = Union[Callable[[Any], None], Callable[[Any, KeyEvent], None]]


class Action:
    def __init__(self, action: str, keys: Optional[Sequence[str]] = None) -> None:
        """
        Action decorator

        Args:
            action: Action
        """
        self.action: str = action
        self.keys: Sequence[str] = keys or []

    def __call__(self, func: ActionHandler) -> ActionHandler:
        setattr(func, "pzp_action", self.action)
        setattr(func, "pzp_keys", self.keys)
        return func


class ActionsHandler:
    def __init__(
        self,
        keys_handler: Optional[KeysHandler] = None,
        keys_binding: Optional[KeysBinding] = None,
    ) -> None:
        """
        Action handler

        Args:
            keys_handler: Keys handler
            keys_binding: Optional[Dict[str, Sequence[str]]] = None,

        Attributes:
            actions: map action names to action handlers
            keys_handler: Keys handler
        """
        self.keys_handler = keys_handler or KeysHandler()
        # Collect the methods with the Action decorator
        self.actions: Dict[str, ActionHandler] = {}
        for name, member in inspect.getmembers(self):
            action: Optional[str] = getattr(member, "pzp_action", None)
            if action is not None:
                self.actions[action] = member
                # Default action
                if name == "default":
                    self.actions["default"] = member
                # Keys binding
                keys: Optional[Sequence[str]] = getattr(member, "pzp_keys", None)
                if keys:
                    self.keys_handler.set_keys_binding(keys, action)
        # Override keys binding
        if keys_binding:
            self.keys_handler.update(keys_binding)

    def process_key_event(self, key_event: KeyEvent) -> None:
        """
        Execute the action hander for a give key event

        Args:
            key_event: Key event to be processed

        Raises:
            MissingHander: If there is no handler for the given key event
        """
        action: str = key_event.action or "default"
        fn = self.actions.get(action)
        if not fn:
            raise MissingHander(action=action, ch=key_event.ch)
        # Check if the function has the key_event argument
        if "key_event" in inspect.getargs(fn.__code__).args:
            fn(key_event=key_event)  # type: ignore
        else:
            fn()  # type: ignore



import pytest
from pzp import pzp
from pzp.screen import Screen
from pzp.exceptions import AcceptAction
from .commons import FakeTerminal


def test_screen_movements():
    terminal = FakeTerminal()
    screen = Screen(stream=terminal.output_stream, fullscreen=True, height=24)
    assert terminal.cursor.y == 0
    screen.move_down(10)
    screen.flush()
    assert terminal.cursor.y == 10
    screen.move_up(5)
    screen.flush()
    assert terminal.cursor.y == 5
    screen.move_right(0)
    screen.flush()
    assert terminal.cursor.x == 0
    screen.move_right(10)
    screen.flush()
    assert terminal.cursor.x == 10
    screen.move_right(100)
    screen.flush()
    assert terminal.cursor.x == 79
    screen.move_left(0)
    screen.flush()
    assert terminal.cursor.x == 79
    screen.move_left(20)
    screen.flush()
    assert terminal.cursor.x == 59
    screen.move_left(200)
    screen.flush()
    assert terminal.cursor.x == 0


def test_screen():
    terminal = FakeTerminal()
    screen = Screen(stream=terminal.output_stream, fullscreen=True, height=24)
    assert terminal.cursor.x == 0
    assert terminal.cursor.y == 0
    screen.space(10).flush()
    assert terminal.cursor.x == 10
    assert terminal.cursor.y == 0
    screen.nl(10).flush()
    assert terminal.cursor.x == 0
    assert terminal.cursor.y == 10
    screen.space().flush()
    assert terminal.cursor.x == 1
    assert terminal.cursor.y == 10
    screen.nl().flush()
    assert terminal.cursor.x == 0
    assert terminal.cursor.y == 11
    screen.space(10).erase_line().flush()
    assert terminal.cursor.x == 10


def test_term():
    terminal = FakeTerminal()
    output_stream = terminal.output_stream
    assert terminal.cursor.y == 0

    lines = 4
    for i in range(0, lines):
        output_stream.write(f"line{i}>\n")
    assert terminal.cursor.y == lines


@pytest.mark.timeout(5)
def test_lazy():
    assert pzp([1], lazy=True) == 1
    assert pzp([], lazy=True) is None
    assert pzp(["a", "b", "c"], input="a", lazy=True) == "a"
    assert pzp(["a", "b", "c"], input="d", lazy=True) is None

    with pytest.raises(AcceptAction) as ex:
        pzp(["a", "b", "c"], input="a", lazy=True, handle_actions=None)
    assert ex.value.selected_item == "a"

    with pytest.raises(AcceptAction) as ex:
        pzp(["a", "b", "c"], input="d", lazy=True, handle_actions=None)
    assert ex.value.selected_item is None



#!/usr/bin/env python

try:
    from .input_posix import get_char, KEYS_MAPPING
except ImportError:
    try:
        from .input_win import get_char, KEYS_MAPPING
    except ImportError:
        raise Exception("Unsupported platform")

__all__ = ["get_char", "KEYS_MAPPING"]



#!/usr/bin/env python

__all__ = ["get_char", "KEYS_MAPPING"]

import sys
import termios
import tty
import os

NULL = "\0"
ESC = "\x1b"
KEYS_MAPPING = {
    "": ESC,
    "[A": "up",
    "[B": "down",
    "[C": "right",
    "[D": "left",
    "[Z": "btab",  # shift-tab
    "[1;2A": "shift-up",
    "[1;2B": "shift-down",
    "[1;2C": "shift-right",
    "[1;2D": "shift-left",
    "[1~": "home",
    "[2~": "insert",
    "[3~": "del",
    "[4~": "end",
    "[5~": "pgup",
    "[6~": "pgdn",
    "OP": "f1",
    "[11~": "f1",
    "OQ": "f2",
    "[12~": "f2",
    "OR": "f3",
    "[13~": "f3",
    "OS": "f4",
    "[14~": "f4",
    "[15~": "f5",
    "[17~": "f6",
    "[18~": "f7",
    "[19~": "f8",
    "[20~": "f9",
    "[21~": "f10",
    "[23~": "f11",
    "[24~": "f12",
}


def get_char() -> str:
    """
    Read a keypress and return the resulting character as a string.

    Returns:
        char: the pressed key or the key description (e.g. "home")
    """
    fd = None
    try:
        fd = sys.stdin.fileno()
        attrs = termios.tcgetattr(fd)
        tty.setraw(fd)
        ch = os.read(fd, 32).decode("utf-8", "replace")
        if ch.startswith(ESC):
            ch = KEYS_MAPPING.get(ch[1:], NULL)
        return ch
    finally:
        if fd is not None:
            termios.tcsetattr(fd, termios.TCSAFLUSH, attrs)



#!/usr/bin/env python

from typing import Sequence, TypeVar

T = TypeVar("T")

__all__ = ["first_or_default"]


def first_or_default(sequence: Sequence[T], default: T) -> T:
    return sequence[0] if sequence else default
